#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['pythonbible',
 'pythonbible.bible',
 'pythonbible.bible.osis',
 'pythonbible.counters']

package_data = \
{'': ['*'],
 'pythonbible': ['.mypy_cache/*',
                 '.mypy_cache/3.9/*',
                 '.mypy_cache/3.9/_typeshed/*',
                 '.mypy_cache/3.9/collections/*',
                 '.mypy_cache/3.9/ctypes/*',
                 '.mypy_cache/3.9/importlib/*',
                 '.mypy_cache/3.9/json/*',
                 '.mypy_cache/3.9/os/*',
                 '.mypy_cache/3.9/pyexpat/*',
                 '.mypy_cache/3.9/pythonbible/*',
                 '.mypy_cache/3.9/pythonbible/bible/*',
                 '.mypy_cache/3.9/pythonbible/bible/osis/*',
                 '.mypy_cache/3.9/pythonbible/counters/*',
                 '.mypy_cache/3.9/xml/*',
                 '.mypy_cache/3.9/xml/etree/*',
                 '.mypy_cache/3.9/xml/parsers/*',
                 '.mypy_cache/3.9/xml/parsers/expat/*',
                 'data/*'],
 'pythonbible.bible': ['data/*', 'data/asv/*', 'data/kjv/*'],
 'pythonbible.bible.osis': ['versions/*']}

extras_require = \
{'all': ['defusedxml >=0.7.1'],
 'dev': ['bandit >=1.7.4',
         'black >=22.1.0',
         'mypy >=0.931',
         'prospector >=1.7.6'],
 'doc': ['furo >=2022.3.4', 'Sphinx >=4.4.0', 'sphinx-toolbox >=2.18.0'],
 'test': ['pytest >=7.0.1', 'pytest-cov >=3.0.0']}

setup(name='pythonbible',
      version='0.7',
      description='pythonbible includes features for parsing texts for scripture references,',
      author='Nathan Patton',
      author_email='npatton@gmail.com',
      url='https://docs.python.bible',
      packages=packages,
      package_data=package_data,
      extras_require=extras_require,
      python_requires='>=3.7',
     )
